@@ -93,6 +93,12 @@ using CoordinateOperationPtr = std::shared_ptr<CoordinateOperation>;
 /** Non-null shared pointer of CoordinateOperation */
 using CoordinateOperationNNPtr = util::nn<CoordinateOperationPtr>;
 
+class Transformation;
+/** Shared pointer of Transformation */
+using TransformationPtr = std::shared_ptr<Transformation>;
+/** Non-null shared pointer of Transformation */
+using TransformationNNPtr = util::nn<TransformationPtr>;
+
 /** \brief Abstract class for a mathematical operation on coordinates.
  *
  * A mathematical operation:
@@ -144,10 +150,12 @@ class PROJ_GCC_DLL CoordinateOperation : public common::ObjectUsage,
 
     /** \brief Return grids needed by an operation. */
     PROJ_DLL virtual std::set<GridDescription>
-    gridsNeeded(const io::DatabaseContextPtr &databaseContext) const = 0;
+    gridsNeeded(const io::DatabaseContextPtr &databaseContext,
+                bool considerKnownGridsAsAvailable) const = 0;
 
     PROJ_DLL bool
-    isPROJInstantiable(const io::DatabaseContextPtr &databaseContext) const;
+    isPROJInstantiable(const io::DatabaseContextPtr &databaseContext,
+                       bool considerKnownGridsAsAvailable) const;
 
     PROJ_DLL bool hasBallparkTransformation() const;
 
@@ -169,12 +177,15 @@ class PROJ_GCC_DLL CoordinateOperation : public common::ObjectUsage,
     PROJ_FRIEND(io::AuthorityFactory);
     PROJ_FRIEND(CoordinateOperationFactory);
     PROJ_FRIEND(ConcatenatedOperation);
+    PROJ_FRIEND(io::WKTParser);
     PROJ_INTERNAL void
     setWeakSourceTargetCRS(std::weak_ptr<crs::CRS> sourceCRSIn,
                            std::weak_ptr<crs::CRS> targetCRSIn);
     PROJ_INTERNAL void setCRSs(const crs::CRSNNPtr &sourceCRSIn,
                                const crs::CRSNNPtr &targetCRSIn,
                                const crs::CRSPtr &interpolationCRSIn);
+    PROJ_INTERNAL void
+    setInterpolationCRS(const crs::CRSPtr &interpolationCRSIn);
     PROJ_INTERNAL void setCRSs(const CoordinateOperation *in,
                                bool inverseSourceTarget);
     PROJ_INTERNAL
@@ -210,7 +221,8 @@ class PROJ_GCC_DLL GeneralOperationParameter : public common::IdentifiedObject {
     PROJ_INTERNAL bool _isEquivalentTo(
         const util::IComparable *other,
         util::IComparable::Criterion criterion =
-            util::IComparable::Criterion::STRICT) const override = 0;
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override = 0;
     //! @endcond
 
   protected:
@@ -251,10 +263,11 @@ class PROJ_GCC_DLL OperationParameter final : public GeneralOperationParameter {
     //! @endcond
 
     //! @cond Doxygen_Suppress
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
     //! @endcond
 
     // non-standard
@@ -307,7 +320,8 @@ class PROJ_GCC_DLL GeneralParameterValue : public util::BaseObject,
     PROJ_INTERNAL bool _isEquivalentTo(
         const util::IComparable *other,
         util::IComparable::Criterion criterion =
-            util::IComparable::Criterion::STRICT) const override = 0;
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override = 0;
     //! @endcond
 
   protected:
@@ -390,10 +404,11 @@ class PROJ_GCC_DLL ParameterValue final : public util::BaseObject,
     PROJ_DLL bool booleanValue() PROJ_PURE_DECL;
 
     //! @cond Doxygen_Suppress
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
     //! @endcond
 
   protected:
@@ -450,10 +465,11 @@ class PROJ_GCC_DLL OperationParameterValue final
     PROJ_INTERNAL void _exportToJSON(io::JSONFormatter *formatter)
         const override; // throw(FormattingException)
 
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
     //! @endcond
 
   protected:
@@ -519,10 +535,11 @@ class PROJ_GCC_DLL OperationMethod : public common::IdentifiedObject,
     PROJ_INTERNAL void _exportToJSON(io::JSONFormatter *formatter)
         const override; // throw(FormattingException)
 
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
     //! @endcond
 
   protected:
@@ -575,18 +592,18 @@ class PROJ_GCC_DLL SingleOperation : virtual public CoordinateOperation {
     PROJ_DLL const OperationMethodNNPtr &method() PROJ_PURE_DECL;
 
     PROJ_DLL const ParameterValuePtr &
-    parameterValue(const std::string &paramName, int epsg_code = 0) const
-        noexcept;
+    parameterValue(const std::string &paramName,
+                   int epsg_code = 0) const noexcept;
 
-    PROJ_DLL const ParameterValuePtr &parameterValue(int epsg_code) const
-        noexcept;
+    PROJ_DLL const ParameterValuePtr &
+    parameterValue(int epsg_code) const noexcept;
 
     PROJ_DLL const common::Measure &
-    parameterValueMeasure(const std::string &paramName, int epsg_code = 0) const
-        noexcept;
+    parameterValueMeasure(const std::string &paramName,
+                          int epsg_code = 0) const noexcept;
 
-    PROJ_DLL const common::Measure &parameterValueMeasure(int epsg_code) const
-        noexcept;
+    PROJ_DLL const common::Measure &
+    parameterValueMeasure(int epsg_code) const noexcept;
 
     PROJ_DLL static SingleOperationNNPtr createPROJBased(
         const util::PropertyMap &properties, const std::string &PROJString,
@@ -595,25 +612,37 @@ class PROJ_GCC_DLL SingleOperation : virtual public CoordinateOperation {
             std::vector<metadata::PositionalAccuracyNNPtr>());
 
     PROJ_DLL std::set<GridDescription>
-    gridsNeeded(const io::DatabaseContextPtr &databaseContext) const override;
+    gridsNeeded(const io::DatabaseContextPtr &databaseContext,
+                bool considerKnownGridsAsAvailable) const override;
 
     PROJ_DLL std::list<std::string> validateParameters() const;
 
+    PROJ_DLL TransformationNNPtr substitutePROJAlternativeGridNames(
+        io::DatabaseContextNNPtr databaseContext) const;
+
     PROJ_PRIVATE :
         //! @cond Doxygen_Suppress
 
         PROJ_DLL double
-        parameterValueNumeric(int epsg_code,
-                              const common::UnitOfMeasure &targetUnit) const
-        noexcept;
+        parameterValueNumeric(
+            int epsg_code,
+            const common::UnitOfMeasure &targetUnit) const noexcept;
 
-    PROJ_INTERNAL double parameterValueNumericAsSI(int epsg_code) const
-        noexcept;
+    PROJ_INTERNAL double parameterValueNumeric(
+        const char *param_name,
+        const common::UnitOfMeasure &targetUnit) const noexcept;
+
+    PROJ_INTERNAL double
+    parameterValueNumericAsSI(int epsg_code) const noexcept;
+
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
+
+    PROJ_INTERNAL bool isLongitudeRotation() const;
 
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
     //! @endcond
 
   protected:
@@ -632,6 +661,7 @@ class PROJ_GCC_DLL SingleOperation : virtual public CoordinateOperation {
 
     PROJ_INTERNAL bool _isEquivalentTo(const util::IComparable *other,
                                        util::IComparable::Criterion criterion,
+                                       const io::DatabaseContextPtr &dbContext,
                                        bool inOtherDirection) const;
 
   private:
@@ -725,7 +755,7 @@ projection grid at the natural origin.
 
 EPSG:8807
 
-\subsection latitude_projection_centre Latitute of projection centre
+\subsection latitude_projection_centre Latitude of projection centre
 
 For an oblique projection, this is the latitude of the point at which the
 azimuth of the central line is defined.
@@ -1310,10 +1340,37 @@ class PROJ_GCC_DLL Conversion : public SingleOperation {
                      const common::Length &falseEasting,
                      const common::Length &falseNorthing);
 
+    PROJ_DLL static ConversionNNPtr
+    createVerticalPerspective(const util::PropertyMap &properties,
+                              const common::Angle &topoOriginLat,
+                              const common::Angle &topoOriginLong,
+                              const common::Length &topoOriginHeight,
+                              const common::Length &viewPointHeight,
+                              const common::Length &falseEasting,
+                              const common::Length &falseNorthing);
+
+    PROJ_DLL static ConversionNNPtr createPoleRotationGRIBConvention(
+        const util::PropertyMap &properties,
+        const common::Angle &southPoleLatInUnrotatedCRS,
+        const common::Angle &southPoleLongInUnrotatedCRS,
+        const common::Angle &axisRotation);
+
+    PROJ_DLL static ConversionNNPtr createPoleRotationNetCDFCFConvention(
+        const util::PropertyMap &properties,
+        const common::Angle &gridNorthPoleLatitude,
+        const common::Angle &gridNorthPoleLongitude,
+        const common::Angle &northPoleGridLongitude);
+
     PROJ_DLL static ConversionNNPtr
     createChangeVerticalUnit(const util::PropertyMap &properties,
                              const common::Scale &factor);
 
+    PROJ_DLL static ConversionNNPtr
+    createChangeVerticalUnit(const util::PropertyMap &properties);
+
+    PROJ_DLL static ConversionNNPtr
+    createHeightDepthReversal(const util::PropertyMap &properties);
+
     PROJ_DLL static ConversionNNPtr createAxisOrderReversal(bool is3D);
 
     PROJ_DLL static ConversionNNPtr
@@ -1339,6 +1396,14 @@ class PROJ_GCC_DLL Conversion : public SingleOperation {
     PROJ_INTERNAL ConversionNNPtr alterParametersLinearUnit(
         const common::UnitOfMeasure &unit, bool convertToNewUnit) const;
 
+    PROJ_INTERNAL static ConversionNNPtr
+    createGeographicGeocentric(const crs::CRSNNPtr &sourceCRS,
+                               const crs::CRSNNPtr &targetCRS);
+
+    PROJ_INTERNAL static ConversionNNPtr
+    createGeographicGeocentricLatitude(const crs::CRSNNPtr &sourceCRS,
+                                       const crs::CRSNNPtr &targetCRS);
+
     //! @endcond
 
   protected:
@@ -1368,12 +1433,6 @@ class PROJ_GCC_DLL Conversion : public SingleOperation {
 
 // ---------------------------------------------------------------------------
 
-class Transformation;
-/** Shared pointer of Transformation */
-using TransformationPtr = std::shared_ptr<Transformation>;
-/** Non-null shared pointer of Transformation */
-using TransformationNNPtr = util::nn<TransformationPtr>;
-
 /** \brief A mathematical operation on coordinates in which parameters are
  * empirically derived from data containing the coordinates of a series of
  * points in both coordinate reference systems.
@@ -1527,9 +1586,6 @@ class PROJ_GCC_DLL Transformation : public SingleOperation {
         const crs::CRSNNPtr &targetCRSIn, const common::Length &offsetHeight,
         const std::vector<metadata::PositionalAccuracyNNPtr> &accuracies);
 
-    PROJ_DLL TransformationNNPtr substitutePROJAlternativeGridNames(
-        io::DatabaseContextNNPtr databaseContext) const;
-
     PROJ_DLL static TransformationNNPtr createChangeVerticalUnit(
         const util::PropertyMap &properties, const crs::CRSNNPtr &sourceCRSIn,
         const crs::CRSNNPtr &targetCRSIn, const common::Scale &factor,
@@ -1545,8 +1601,6 @@ class PROJ_GCC_DLL Transformation : public SingleOperation {
 
     PROJ_INTERNAL const std::string &getHeightToGeographic3DFilename() const;
 
-    PROJ_INTERNAL bool isLongitudeRotation() const;
-
     PROJ_INTERNAL void _exportToWKT(io::WKTFormatter *formatter)
         const override; // throw(io::FormattingException)
 
@@ -1555,6 +1609,14 @@ class PROJ_GCC_DLL Transformation : public SingleOperation {
 
     PROJ_INTERNAL TransformationNNPtr shallowClone() const;
 
+    PROJ_INTERNAL TransformationNNPtr
+    promoteTo3D(const std::string &newName,
+                const io::DatabaseContextPtr &dbContext) const;
+
+    PROJ_INTERNAL TransformationNNPtr
+    demoteTo2D(const std::string &newName,
+               const io::DatabaseContextPtr &dbContext) const;
+
     //! @endcond
 
   protected:
@@ -1570,6 +1632,8 @@ class PROJ_GCC_DLL Transformation : public SingleOperation {
     PROJ_INTERNAL void _exportToPROJString(io::PROJStringFormatter *formatter)
         const override; // throw(FormattingException)
 
+    PROJ_FRIEND(CoordinateOperationFactory);
+    PROJ_FRIEND(SingleOperation);
     PROJ_INTERNAL TransformationNNPtr inverseAsTransformation() const;
 
     PROJ_INTERNAL CoordinateOperationNNPtr _shallowClone() const override;
@@ -1644,7 +1708,8 @@ class PROJ_GCC_DLL ConcatenatedOperation final : public CoordinateOperation {
         bool checkExtent); // throw InvalidOperation
 
     PROJ_DLL std::set<GridDescription>
-    gridsNeeded(const io::DatabaseContextPtr &databaseContext) const override;
+    gridsNeeded(const io::DatabaseContextPtr &databaseContext,
+                bool considerKnownGridsAsAvailable) const override;
 
     PROJ_PRIVATE :
 
@@ -1653,10 +1718,11 @@ class PROJ_GCC_DLL ConcatenatedOperation final : public CoordinateOperation {
         _exportToWKT(io::WKTFormatter *formatter)
             const override; // throw(io::FormattingException)
 
-    PROJ_INTERNAL bool
-    _isEquivalentTo(const util::IComparable *other,
-                    util::IComparable::Criterion criterion =
-                        util::IComparable::Criterion::STRICT) const override;
+    PROJ_INTERNAL bool _isEquivalentTo(
+        const util::IComparable *other,
+        util::IComparable::Criterion criterion =
+            util::IComparable::Criterion::STRICT,
+        const io::DatabaseContextPtr &dbContext = nullptr) const override;
 
     PROJ_INTERNAL void _exportToJSON(io::JSONFormatter *formatter)
         const override; // throw(FormattingException)
@@ -1718,6 +1784,10 @@ class PROJ_GCC_DLL CoordinateOperationContext {
 
     PROJ_DLL void setDesiredAccuracy(double accuracy);
 
+    PROJ_DLL void setAllowBallparkTransformations(bool allow);
+
+    PROJ_DLL bool getAllowBallparkTransformations() const;
+
     /** Specify how source and target CRS extent should be used to restrict
      * candidate operations (only taken into account if no explicit area of
      * interest is specified. */
@@ -1773,6 +1843,12 @@ class PROJ_GCC_DLL CoordinateOperationContext {
         /** Ignore grid availability at all. Results will be presented as if
          * all grids were available. */
         IGNORE_GRID_AVAILABILITY,
+
+        /** Results will be presented as if grids known to PROJ (that is
+         * registered in the grid_alternatives table of its database) were
+         * available. Used typically when networking is enabled.
+         */
+        KNOWN_AVAILABLE,
     };
 
     PROJ_DLL void setGridAvailabilityUse(GridAvailabilityUse use);
